class Magic
  #DefaultMagicDB = "/etc/magic"
  DefaultMagicDB = "./magic"
  BlockSize = 512

  class Pattern
    include Comparable

    attr_reader :offset
    def initialize(offset, type, value = "", message = "")
      @offset = offset.to_i
      if type =~ /string(?:\/([bBc]))?/
	@type = "string"
	@operator = $1 if $1
	if value == ">\\0"
	  @value = /[^\000\n]+$/	# match any to `\n'
	else
	  @value = /#{Regexp.quote(eval("%\001#{value}\001"))}/
	end
      else
	if value =~ /^[<>=&^x]/
	  op = $&
	  value = $'
	end
	case op
	when /[<>&^x]/
	  @operator = op
	else # op == "=" or none
	  @operator = "=="
	end
	@value = (Integer(value) rescue 0)

	mask = nil
	if type =~ /&(.+)/
	  type = $`
	  mask = Integer($1)
	end
	@type = type
	@value &= mask unless mask.nil?
      end
      @matcher = "match_"+@type
      if not respond_to? @matcher
	@matcher = "match_unknown"
      end
      @message = message
      @sub = []
    end

    def match(str)
      if m = send(@matcher, str)
	m[0].gsub!(/%[hl]([diouXx])/) { "%#{$1}" }
	mesg = sprintf(*m)
	@sub.sort.each do
	  |pat|
	  m, v = pat.match(str)
	  if m
	    m.gsub!(/%[hl]([diouXx])/) { "%#{$1}" }
	    mesg << sprintf(m, v)
	  end
	end
	mesg
      else
	nil
      end
    end

    def match_string(str)
      idx = str.index(@value, @offset)
      [@message, $&] if idx == @offset
    end

    def match_byte(str)
      byte = str[@offset, 1].unpack("c").shift
      byte += 0xff if @type =~ /^u/ and byte < 0
      [@message, byte] if eval_with(byte)
    end
    alias match_ubyte match_byte
    alias match_char match_byte
    alias match_uchar match_byte

    def match_short(str)
      case @type
      when /^be/
	t = "n"
      when /^le/
	t = "v"
      else
	t = "s"
      end
      short = str[@offset, 2].unpack(t).shift
      short += 0xffff if @type =~ /^u/ and short < 0
      [@message, short] if eval_with(short)
    end
    alias match_beshort match_short
    alias match_leshort match_short
    alias match_ushort match_short
    alias match_ubeshort match_short
    alias match_uleshort match_short

    def match_long(str)
      case @type
      when /^be/
	t = "N"
      when /^le/
	t = "V"
      else
	t = "l"
      end
      long = str[@offset, 4].unpack(t).shift
      long += 0xffffffff if @type =~ /^u/ and long < 0
      [@message, long] if eval_with(long)
    end
    alias match_belong match_long
    alias match_lelong match_long
    alias match_ulong match_long
    alias match_ubelong match_long
    alias match_ulelong match_long
    alias match_date match_long
    alias match_bedate match_long
    alias match_ledate match_long

    def match_unknown(str)
      @message			# always match
    end

    def eval_with(n)
      case @operator
      when "x"
	true
      when "<", ">", "=="
	n.send(@operator, @value)
      when "&", "^"
	n.send(@operator, @value) == @value
      else
	nil
      end
    end

    def <<(pat)
      @sub << pat
    end

    def <=>(o)
      @offset <=> o.offset
    end

    def inspect
      "#{@offset}:#{@type}:#{@matcher}:#{@value.inspect}:#{@message}"+
      (@sub.empty? ? "" :
       "\n"+@sub.collect {|s| s.inspect}.join("\n").gsub(/^/, "  "))
    end
  end

  def initialize(db = DefaultMagicDB)
    @db = []
    File.open(db) do
      |inf|
      cur = nil
      clev = 0
      stk = []
      inf.each do
	|ln|
	ln.sub!(/^#.*$/, "")
	next if ln =~ /^\s*$/

	lev = ln.index(/[^>]/)
	next if lev.nil?	# ignore corrupted line
	ln[0, lev] = ""
	ln.gsub!(/\\([ \t\n])/) { "\\%03o"%$1[0] }
	offset, type, value, message = ln.split(/\s+/, 4)
	value = "" unless value
	message = "" unless message
	if lev == 0
	  clev = 0
	  cur = Pattern.new(offset, type, value, message)
	  stk[clev] = cur
	  @db << cur
	else
	  if lev != clev
	    stk[clev] = cur
	    clev = lev
	  end
	  cur = Pattern.new(offset, type, value, message)
	  stk[clev-1] << cur
	end
      end
    end
    @db.sort!
  end

  def match(file)
    mesg = nil
    File.open(file) do
      |inf|
      blk = inf.read(BlockSize)
      catch :break do
	@db.each do
	  |pat|
	  while pat.offset > blk.size
	    add = inf.read(BlockSize)
	    throw :break if add.nil?
	    blk << add
	  end

	  mesg = pat.match(blk)
	  break if mesg
	end
      end
    end
    mesg.chomp.gsub(/\n/, " ").gsub(/.\\b/, "") if mesg
  end

  def inspect
    @db.collect { |e| e.inspect }.join("\n")
  end
end


if $0 == __FILE__

file = ARGV.shift
magic = ARGV.shift

if file.nil?
  $stderr.puts "Usage: #{$0} file [magic]"
  exit(0)
end

if magic
  m = Magic.new(magic)
else
  m = Magic.new
end
p m.match(file)

end
